-- fed.ex : font editor

-- Jiri Babor
-- j.babor@gns.cri.nz
-- version 1.00   96-10-17
-- version 1.01   97-02-16
--    new font routine now loads maximum blank set - 256 characters
--    paper color (fpc) now detected
--    cosmetic changes

constant
   on = 1,
   true=1,
   false=0,
   off = 0,
   x0 = 10,
   y0 = 180,
   xt = 10,
   yt = 85,
   xb = 65,
   yb = 85,
   xc=303,
   yc=10,
   esc=27

type bool(integer x)
   return x=0 or x=1
end type

sequence
   font_file,
   b,             -- char bytes
   b_b,           -- char bytes back up
   w,             -- char widths
   w_b,           -- char widths backup
   mb,            -- mouse boxes, checked by mouse_box function
   char_mat,      -- char matrix
   buff_mat,      -- buffer char matrix
   ms             -- polled mouse status - see function poll_mouse() in rat.e

integer
   cch,           -- current char
   key,
   simple09,
   simple12,
   thin8x10,
   maxw,
   maxw_buff,
   maxh,
   maxh_buff,
   w_buff,
   fc,
   lc,
   byte,
   blue_line,
   mx,my,
   mx0,my0,
   fdc,           -- first display char
   ldc,           -- last display char
   dcc,           -- display column count
   drc            -- display row count


bool buff_loaded, loaded, grid_flag

include graphics.e
include get.e
include font.e
include gadgets.e
include mouse.e
include selector.e

function div(integer a,integer b)
   return  floor(a/b)
end function   -- div

function mod(integer a,integer b)
   return remainder(a,b)
end function   -- mod

procedure load_font()
   -- load font file into memory
   integer fn
   sequence s
   fn = open(font_file, "rb")
   if fn = -1 then
      puts(1, "Font load error: " & font_file)
      abort(1)
   else
      fc = getc(fn)
      lc = getc(fn)
      maxw = getc(fn)
      maxh = getc(fn)
      byte = getc(fn)   -- junk
      byte = getc(fn)   -- junk
      w = {}
      b = {}
      for i = fc to lc do
         w = w & getc(fn)
      end for
      for i = fc to lc do
         s = {}
         for j = 1 to maxw*maxh do
            s = s & getc(fn)
         end for
         b = append(b,s)
      end for
      close(fn)
      -- back up
      w_b = w
      b_b = b
   end if
end procedure  -- load_font

procedure save_char()
   sequence s
   for j = 1 to maxh do
      for i = 1 to maxw do
         s={}
         for k = 1 to 8 do
            s=char_mat[j][(i-1)*8+k]&s
         end for
         b[cch-fc+1][maxw*(j-1)+i] = bits_to_int(s)
      end for
   end for
end procedure  -- save_char

procedure save_font()
   -- save current font file to disk
   integer fn
   fn = open(font_file, "wb")
   if fn = -1 then
      puts(1, "Font save error: " & font_file)
      abort(1)
   else
      save_char()
      puts(fn,fc)
      puts(fn,lc)
      puts(fn,maxw)
      puts(fn,maxh)
      puts(fn,0)
      puts(fn,0)
      for i = 1 to lc-fc+1 do
         puts(fn,w[i])
      end for
      for i = 1 to lc-fc+1 do
         for j = 1 to maxw*maxh do
            puts(fn,b[i][j])
         end for
      end for
      close(fn)
   end if
end procedure  -- save_font

procedure show_badge()
  cs={12,4,0}
  frame({10,10},{153,80})
  cs={0,4,12}
  frame({13,13},{150,77})
  fselect(simple12)
  ftc=15
  fsc=0
  fp[1]=on
  fy=20
  fcprint(10,153,"font editor")
  fselect(simple09)
  ftc=14
  fy=40
  fcprint(10,153,"ver 1.01")
  fy=60
  fcprint(10,153,"jiri babor  1996")
end procedure  -- show_badge

procedure show_buttons()
   integer h,w,x,y
   sequence label
   fselect(simple09)
   cs={15,7,8}
   fp[1]=0
   ftc=1
   fsc=8
   x=158
   y=10
   w=70
   h=17
   label={{"new","load"},
          {"save","save as"},
          {"prev char","next char"},
          {"prepend","append"},
          {"to buffer","from buff"},
          {"restore","delete"}}
   for j=1 to 2 do
      for i = 1 to 6 do
         button({x+(j-1)*w,y+(i-1)*h},{x+j*w-1,y+i*h-1},label[i][j])
      end for
   end for
   button({158,112},{203,128},"grid")
   button({204,112},{250,128},"wipe")
   button({251,112},{297,128},"quit")
   cs={15,4,8}
   fp[1]=false
   ftc=15
   button({158,132},{174,149},{17})
   button({175,132},{192,149},{16})
   button({193,132},{209,149},{30})
   button({210,132},{227,149},{31})
   ftc=1
   ftc=7
   button({228,132},{244,149},{17})
   button({245,132},{262,149},{16})
   ftc=9
   button({263,132},{279,149},{30})
   button({280,132},{297,149},{31})
end procedure  -- show_buttons

procedure show_font_details()
   -- max width in pixels
   rect(1,1,{57,159},{76,171})
   fp[1]=on
   ftc=14
   fsc=1
   fx=63
   fy=160
   fprint(sprintf("%2d",maxw*8))
   -- max height in pixels
   rect(1,1,{130,159},{148,171})
   ftc=14
   fx=136
   fy=160
   fprint(sprintf("%2d",maxh))
   -- font file name
   rect(1,1,{190,159},{292,171})
   ftc=14
   fx=196
   fy=160
   fprint(font_file)
end procedure  -- show_font_details

procedure show_file_boxes()
   -- size boxes
   cs={15,7,8}
   frame({ 10,154},{ 80,175})
   frame({ 83,154},{153,175})
   cs={8,1,15}
   frame({ 55,157},{ 77,172})
   frame({128,157},{150,172})
   fselect(simple09)
   ftc=0
   fp[1]=off
   fx=15
   fy=160
   fprint("width")
   fx=88
   fprint("height")
   -- filename box
   cs={15,7,8}
   frame({158,154},{297,175})
   cs={8,1,15}
   frame({188,157},{294,172})
   fselect(simple09)
   ftc=0
   fp[1]=off
   fx=164
   fprint("file")
end procedure  -- show_file_boxes

procedure change_cell()
   integer row, col
   col = div(mouse_event[2]-x0,6)+1
   row = div(mouse_event[3]-y0,6)+1
   if col<=maxw*8 and row<=maxh then
      char_mat[row][col] = not char_mat[row][col]
      if char_mat[row][col] then
         rect(15,1,{x0+(col-1)*6+1,y0+(row-1)*6+1},{x0+col*6-1,y0+row*6-1})
         pixel(15,{xt+col-1,yt+row-1})
      elsif col<=w[cch-fc+1] then
         rect(0,1,{x0+(col-1)*6+1,y0+(row-1)*6+1},{x0+col*6-1,y0+row*6-1})
         pixel(0,{xt+col-1,yt+row-1})
      else
         rect(8,1,{x0+(col-1)*6+1,y0+(row-1)*6+1},{x0+col*6-1,y0+row*6-1})
         pixel(8,{xt+col-1,yt+row-1})
      end if
   end if
end procedure  -- change_cell

procedure show_char_space()
   rect(8,0,{x0,y0},{x0+6*8*maxw,y0+6*maxh})
   if w[cch-fc+1] then
      rect(0,1,{x0+1,y0+1},{x0+w[cch-fc+1]*6-1,y0+6*maxh-1})
      if grid_flag then
         for i = 1 to w[cch-fc+1]-1 do
            draw_line(8,{{x0+i*6,y0+1},{x0+i*6,y0+6*maxh-1}})
         end for
         for i = 1 to maxh-1 do
            if i=blue_line then
               draw_line(1,{{x0+1,y0+i*6},{x0+6*w[cch-fc+1]-1,y0+i*6}})
            else
               draw_line(8,{{x0+1,y0+i*6},{x0+6*w[cch-fc+1]-1,y0+i*6}})
            end if
         end for
      end if
   end if
   if w[cch-fc+1]<maxw*8 then
      rect(8,1,{x0+w[cch-fc+1]*6,y0+1},{x0+6*8*maxw,y0+6*maxh-1})
      if grid_flag then
         for i = w[cch-fc+1]+1 to maxw*8-1 do
            draw_line(7,{{x0+i*6,y0+1},{x0+i*6,y0+maxh*6-1}})
         end for
         for i = 1 to maxh-1 do
            if i=blue_line then
               draw_line(9,{{x0+w[cch-fc+1]*6+1,y0+i*6},{x0+6*8*maxw-1,y0+i*6}})
            else
               draw_line(7,{{x0+w[cch-fc+1]*6+1,y0+i*6},{x0+6*8*maxw-1,y0+i*6}})
            end if
         end for
      end if
   end if
end procedure  -- show_char_space

procedure load_char_mat()
   sequence s
   char_mat = repeat(repeat(0,maxw*8),maxh)
   for i = 1 to maxh do
      for j = 1 to maxw do
         byte = b[cch-fc+1][maxw*(i-1)+j]
         if byte then
            s = int_to_bits(byte,8)
            for k = 1 to 8 do
               char_mat[i][(j-1)*8+9-k] = s[k]
            end for
         end if
      end for
   end for
end procedure  -- load_char_mat

procedure show_char()
   for i = 1 to maxh do
      for j = 1 to maxw*8 do
         if char_mat[i][j] then
            rect(15,1,{x0+(j-1)*6+1,y0+(i-1)*6+1},{x0+j*6-1,y0+i*6-1})
         end if
      end for
   end for
end procedure  -- show_char

procedure show_buffer_tile()
   rect(12,1,{xb,yb},{xb+47,yb+47})
   if loaded then
      rect(8,1,{xb,yb},{xb+8*maxw_buff-1,yb+maxh_buff-1})
      if w_buff>0 then
         rect(0,1,{xb,yb},{xb+w_buff-1,yb+maxh_buff-1})
      end if
      for i = 1 to maxh_buff do
         for j = 1 to maxw_buff*8 do
            if buff_mat[i][j] then
               pixel(15,{xb+j-1,yb+i-1})
            end if
         end for
      end for
   else
      fselect(simple09)
      ftc=14
      fsc=8
      fp[1]=true
      fx = xb
      fy=yb+52
      fprint("buffer")
   end if
end procedure  -- show_buffer_tile

procedure show_char_tile()
   rect(12,1,{xt,yt},{xt+47,yt+47})
   if loaded then
      rect(8,1,{xt,yt},{xt+8*maxw-1,yt+maxh-1})
      if w[cch-fc+1] then
         rect(0,1,{xt,yt},{xt+w[cch-fc+1]-1,yt+maxh-1})
      end if
      for i = 1 to maxh do
         for j = 1 to maxw*8 do
            if char_mat[i][j] then
               pixel(15,{xt+j-1,yt+i-1})
            end if
         end for
      end for
      -- show ascii code
      rect(7,1,{xt+30,yt+52},{xt+47,yt+62})
      fselect(simple09)
      ftc=1
      fp[1]=false
      fx=xt+30
      fy=yt+53
      fprint(sprintf("%d",cch))
   else
      fselect(simple09)
      ftc=14
      fsc=8
      fp[1]=true
      fx = xt
      fy=yt+52
      fprint("ascii")
   end if
end procedure  -- show_char_tile

procedure show_table()
   rect(12,1,{x0,y0},{x0+6*48,y0+6*48})   -- red table
end procedure  -- show_table

procedure load()
   object ff
   ff = selected(302,10,"*.f",0)
   if sequence(ff) then
      font_file=ff
      load_font()
      loaded=true
      if cch<fc then
         cch=fc
      elsif cch>lc then
         cch=lc
      end if
      load_char_mat()

      fdc=fc                     -- first display char
      dcc=div(639-xc,maxw*8+1)   -- display column count
      drc=div(479-yc,maxh+1)     -- display row count
      if blue_line>maxh  then
         blue_line=maxh-2
      end if
      show_table()
      show_char_space()
      show_char()
      show_char_tile()
      show_font_details()
      mb[24]={{x0,y0},{x0+maxw*8*6-1,y0+maxh*6-1}}
   end if
end procedure  -- load

procedure new_font()
   sequence s,v
   fselect(simple09)
   rect(1,1,{57,159},{75,171})
   ftc=14
   fpc=-1
   fp[1]=off
   fx=63
   fy=160
   s={}
   s=fgets(s)
   v=value(s)
   if v[1]=GET_SUCCESS then
      maxw=floor(v[2]/8)
      if remainder(v[2],8)!=0 then
         maxw=maxw+1
      end if
      if maxw<1 then
         maxw=1
      elsif maxw>6 then
         maxw=6
      end if
      rect(1,1,{57,159},{76,171})
      fx=63
      fprint(sprintf("%d",maxw*8))
   else
      puts(1,"Aborted! Incorrect value entered...")
      abort(1)
   end if
   rect(1,1,{130,159},{148,171})
   fx=136
   s={}
   s=fgets(s)
   v=value(s)
   if v[1]=GET_SUCCESS then
      maxh=v[2]
      if maxh<2 then
         maxh=2
      elsif maxh>48 then
         maxh=48
      end if
   else
      puts(1,"Aborted! Incorrect value entered...")
      abort(1)
   end if
   rect(1,1,{190,159},{292,171})
   fx=196
   font_file=fgets("")
   w=repeat(0,256)
   w_b=w
   b=repeat(repeat(0,maxw*maxh),256)
   b_b=b
   fc=0
   lc=255
   load_char_mat()
   loaded=true

   fdc=0                      -- first display char
   dcc=div(639-xc,maxw*8+1)   -- display column count
   drc=div(479-yc,maxh+1)     -- display row count

   if blue_line>maxh  then
      blue_line=maxh-2
   end if
   show_table()
   show_char_space()
   show_char()
   show_char_tile()
   show_font_details()
   mb[24]={{x0,y0},{x0+maxw*8*6-1,y0+maxh*6-1}}
end procedure  -- new_font

procedure paste()
   integer  mw,mh
   w[cch-fc+1]=w_buff
   if w[cch-fc+1]>maxw*8 then
      w[cch-fc+1]=maxw*8
   end if
   if maxw_buff>maxw then
      mw=maxw
   else
      mw=maxw_buff
   end if
   if maxh_buff>maxh then
      mh=maxh
   else
      mh=maxh_buff
   end if
   for i = 1 to mh do
      char_mat[i][1..mw*8]=buff_mat[i][1..mw*8]
   end for
end procedure  -- paste

procedure save_as()
   sequence s
   fselect(simple09)
   s=font_file
   rect(1,1,{190,159},{292,171})
   fpc=-1
   ftc=14
   fx=196
   fy=160
   font_file=fgets(s)
   save_font()
end procedure  -- save_as

procedure print_char(integer c)
   integer color,x,y
   sequence s
   if c=cch then
      color=14
   else
      color=15
   end if
   x=xc+mod(c-fdc,dcc)*(8*maxw+1)
   y=yc+div(c-fdc,dcc)*(maxh+1)
   if w[c-fc+1] then
      rect(0,1,{x,y},{x+w[c-fc+1]-1,y+maxh-1})
   end if
   if w[c-fc+1]<maxw*8 then
      rect(8,1,{x+w[c-fc+1],y},{x+8*maxw-1,y+maxh-1})
   end if
   for i = 1 to maxh do
      for j = 1 to maxw do
         byte = b[c-fc+1][maxw*(i-1)+j]
         if byte then
            s = int_to_bits(byte,8)
            for k = 1 to 8 do
               if s[k] then
                  pixel(color,{x+(j-1)*8+8-k,y+i-1})
               end if
            end for
         end if
      end for
   end for
end procedure  -- print_char

procedure pick_char()
   integer c,oc
   c = fdc + div(mouse_event[2]-xc,maxw*8+1) + div(mouse_event[3]-yc,maxh+1)*dcc
   if c<=lc then
      save_char()
      oc=cch
      cch=c
      print_char(oc)
      print_char(cch)
      load_char_mat()
      show_table()
      show_char_space()
      show_char()
      show_char_tile()
   end if
end procedure  -- pick_char

procedure display_font()
   ldc=fdc+dcc*drc-1
   if cch<fdc then
      while cch<fdc do
         fdc=fdc-dcc*drc
      end while
      ldc=fdc+dcc*drc-1
   elsif cch>ldc then
      while cch>ldc do
         ldc=ldc+dcc*drc
      end while
      fdc=ldc-dcc*drc+1
   end if
   if fdc<fc then
      fdc=fc
      ldc=fdc+dcc*drc-1
   end if
   if ldc>lc then
      ldc=lc
   end if
   -- erase old display
   rect(7,1,{xc,yc},{639,479})
   for i = fdc to ldc do
      print_char(i)
   end for
end procedure  -- display_font

procedure prev_char()
   if cch>fc then
      bounce({158,44},{227,60})
      save_char()
      cch=cch-1
      if cch<fdc then
         display_font()
      else
         print_char(cch)
         print_char(cch+1)
      end if
      load_char_mat()
      show_table()
      show_char_space()
      show_char()
      show_char_tile()
   end if
end procedure  -- prev_char

procedure next_char()
   if cch<lc then
      bounce({228,44},{297,60})
      save_char()
      cch=cch+1
      if cch>ldc then
         display_font()
      else
         print_char(cch)
         print_char(cch-1)
      end if
      show_table()
      load_char_mat()
      show_char_space()
      show_char()
      show_char_tile()
   end if
end procedure  -- next_char

procedure prepend_char()
   if fc>0 then
      bounce({158,61},{227,77})
      fc=fc-1
      b=prepend(b,repeat(0,maxw*maxh))
      b_b=b
      w=prepend(w,0)
      w_b=w
      cch=fc
      load_char_mat()
      show_table()
      show_char_space()
      show_char()
      show_char_tile()
      fdc=fc
      display_font()
   end if
end procedure  -- prepend_char

procedure append_char()
   if lc<255 then
      bounce({228,61},{297,77})
      lc=lc+1
      b=append(b,repeat(0,maxw*maxh))
      b_b=b
      w=append(w,0)
      w_b=w
      cch=lc
      load_char_mat()
      show_table()
      show_char_space()
      show_char()
      show_char_tile()
      display_font()
   end if
end procedure  -- append_char

procedure del_char()
   if cch=fc or cch=lc then
      bounce({228,95},{297,111})
      if cch=fc then
         b=b[2..length(b)]
         b_b=b
         w=w[2..length(w)]
         w_b=w
         cch=cch+1
         fc=cch
         fdc=fc
      else
         b=b[1..length(b)-1]
         b_b=b
         w=w[1..length(w)-1]
         w_b=w
         cch=cch-1
         lc=cch
      end if
      load_char_mat()
      show_table()
      show_char_space()
      show_char()
      show_char_tile()
      display_font()
   end if
end procedure  -- del_char

-- main ------------------------------------------------------------------------

simple09 = fload("simple09.f")      -- display fonts
simple12 = fload("simple12.f")
thin8x10 = fload("thin8x10.f")
mouse_events(LEFT_DOWN+RIGHT_DOWN)  -- first call - it turns mouse on
mouse_pointer(0)                    -- turn mouse pointer off again

--key=wait_key()

if graphics_mode(18) then
   puts(1, "Need VGA graphics!\n")
   abort(1)
end if

-- clear screen
rect(7,1,{0,0},{639,479})

-- initial defaults
grid_flag = on
loaded = false
buff_loaded = false
cch='A'
blue_line = 48

show_badge()
show_buttons()
show_table()
show_char_tile()
show_buffer_tile()
show_file_boxes()
-- show boxes for rows & columns
cs={9,1,0}
frame({117,85},{153,115})
frame({117,119},{153,149})
fselect(simple09)
fp[1]=true
ftc=14
fsc=0
fy=103
fcprint(117,153,"col")
fy=137
fcprint(117,153,"row")

mb={}
for i = 1 to 6 do
   for j=1 to 2 do
      mb= append(mb,{{158+(j-1)*70,10+(i-1)*17},{158+j*70-1,10+i*17-1}})
   end for
end for
mb=append(mb,{{158,112},{203,128}}) -- 13: grid
mb=append(mb,{{204,112},{250,128}}) -- 14: wipe
mb=append(mb,{{251,112},{297,128}}) -- 15: quit
mb=append(mb,{{158,132},{174,149}}) -- 16: shift left
mb=append(mb,{{175,132},{192,149}}) -- 17: shift right
mb=append(mb,{{193,132},{209,149}}) -- 18: shift up
mb=append(mb,{{210,132},{227,149}}) -- 19: shift down
mb=append(mb,{{228,132},{244,149}}) -- 20: narrower char
mb=append(mb,{{245,132},{262,149}}) -- 21: wider char
mb=append(mb,{{263,132},{279,149}}) -- 22: blue line up
mb=append(mb,{{280,132},{297,149}}) -- 23: blue line down
mb=append(mb,{{ x0, y0},{ x0, y0}}) -- 24: dummy char space - until loaded
mb=append(mb,{{xc,yc},{639,479}})   -- 25: font display area

-- main loop
mx0=0 my0=0    -- init mouse status, for later comparisons
while true do
   mouse_pointer(1)
   while true do
      if loaded then
         ms = poll_mouse()
         mx = div(ms[2]-x0,6)+1
         my = div(ms[3]-y0,6)+1
         if mx>0 and mx<=maxw*8 and my>0 and my<=maxh and (mx!=mx0 or my!=my0) then
            mx0=mx my0=my
            rect(9,1,{120, 88},{150,102})
            rect(9,1,{120,122},{150,136})
            fselect(thin8x10)
            fp[1]=off
            ftc=15
            fy=91
            fcprint(117,153,sprintf("%2d",mx))
            fy=125
            fcprint(117,153,sprintf("%2d",my))
         end if
      end if
      mouse_event = get_mouse()
      key = get_key()
      if sequence(mouse_event) or key!=-1 then
         exit
      end if
   end while
   mouse_pointer(0)
   if sequence(mouse_event) then
      if mouse_event[1]=LEFT_DOWN then
         if mouse_box(mb) then
            if mbox=1 then                    -- new font
               bounce({158,10},{227,26})
               new_font()
               display_font()
            elsif mbox=2 then                 -- load font
               bounce({228,10},{297,26})
               load()
               if loaded then
                  display_font()
               end if
            elsif mbox=3 and loaded then      -- save font
               bounce({158,27},{227,43})
               save_font()
            elsif mbox=4 and loaded then      -- save as
               bounce({228,27},{297,43})
               save_as()
            elsif mbox=5 and loaded then      -- prev char
               prev_char()
            elsif mbox=6 and loaded then      -- next_char
               next_char()
            elsif mbox=7 and loaded then      -- prepend char
               prepend_char()
            elsif mbox=8 and loaded then      -- append char
               append_char()
            elsif mbox=9 and loaded then      -- copy char to buffer
               bounce({158,78},{227,94})
               buff_mat=char_mat
               w_buff=w[cch-fc+1]
               maxw_buff=maxw
               maxh_buff=maxh
               show_buffer_tile()
               buff_loaded=true
            elsif mbox=10 and loaded then     -- paste char back from buffer
               if buff_loaded then
                  bounce({228,78},{297,94})
                  paste()
                  show_table()
                  show_char_space()
                  show_char()
                  show_char_tile()
               end if
            elsif mbox=11 and loaded then     -- restore char
               bounce({158,95},{227,111})
               b[cch-fc+1]=b_b[cch-fc+1]
               w[cch-fc+1]=w_b[cch-fc+1]
               load_char_mat()
               show_table()
               show_char_space()
               show_char()
               show_char_tile()
            elsif mbox=12 and loaded then     -- delete first or last char
               del_char()
            elsif mbox=13 and loaded then     -- grid toggle
               bounce({158,112},{203,128})
               grid_flag=not grid_flag
               show_char_space()
               show_char()
            elsif mbox=14 and loaded then     -- wipe char
               bounce({204,112},{250,128})
               char_mat = repeat(repeat(0,maxw*8),maxh)
               --load_char_mat()
               show_table()
               show_char_space()
               show_char()
               show_char_tile()
            elsif mbox=15 then                -- quit
               exit
            elsif mbox=16 and loaded then     -- shift char left
               for i = 1 to maxh do
                  char_mat[i]=char_mat[i][2..length(char_mat[i])]
                  char_mat[i]=append(char_mat[i],0)
               end for
                  show_char_space()
                  show_char()
                  show_char_tile()
            elsif mbox=17 and loaded then     -- shift char right
               for i = 1 to maxh do
                  char_mat[i]=char_mat[i][1..length(char_mat[i])-1]
                  char_mat[i]=prepend(char_mat[i],0)
               end for
                  show_char_space()
                  show_char()
                  show_char_tile()
            elsif mbox=18 and loaded then  -- shift char up
               char_mat=char_mat[2..length(char_mat)]
               char_mat=append(char_mat,repeat(0,maxw*8))
                  show_char_space()
                  show_char()
                  show_char_tile()
            elsif mbox=19 and loaded then  -- shift char down
               char_mat=char_mat[1..length(char_mat)-1]
               char_mat=prepend(char_mat,repeat(0,maxw*8))
                  show_char_space()
                  show_char()
                  show_char_tile()
            elsif mbox=20 and loaded then  -- narrower char
               if w[cch-fc+1]>0 then
                  w[cch-fc+1]=w[cch-fc+1]-1
                  show_char_space()
                  show_char()
                  show_char_tile()
               end if
            elsif mbox=21 and loaded then  -- wider char
               if w[cch-fc+1]<maxw*8 then
                  w[cch-fc+1]=w[cch-fc+1]+1
                  show_char_space()
                  show_char()
                  show_char_tile()
               end if
            elsif mbox=22 and loaded then  -- blue line up
               if blue_line>1 and grid_flag then
                  if w[cch-fc+1]>0 then
                     draw_line(8,{{x0,y0+blue_line*6},{x0+6*w[cch-fc+1]-1,y0+blue_line*6}})
                  end if
                  if w[cch-fc+1]<maxw*8 then
                     draw_line(7,{{x0+w[cch-fc+1]*6+1,y0+blue_line*6},{x0+6*8*maxw-1,y0+blue_line*6}})
                  end if
                  blue_line = blue_line-1
                  if w[cch-fc+1]>0 then
                     draw_line(1,{{x0,y0+blue_line*6},{x0+6*w[cch-fc+1]-1,y0+blue_line*6}})
                  end if
                  if w[cch-fc+1]<maxw*8 then
                     draw_line(9,{{x0+w[cch-fc+1]*6+1,y0+blue_line*6},{x0+6*8*maxw-1,y0+blue_line*6}})
                  end if
               end if
            elsif mbox=23 and loaded then  -- blue line down
               if blue_line<maxh-1 and grid_flag then
                  if w[cch-fc+1]>0 then
                     draw_line(8,{{x0,y0+blue_line*6},{x0+6*w[cch-fc+1]-1,y0+blue_line*6}})
                  end if
                  if w[cch-fc+1]<maxw*8 then
                     draw_line(7,{{x0+w[cch-fc+1]*6+1,y0+blue_line*6},{x0+6*8*maxw-1,y0+blue_line*6}})
                  end if
                  blue_line = blue_line+1
                  if w[cch-fc+1]>0 then
                     draw_line(1,{{x0,y0+blue_line*6},{x0+6*w[cch-fc+1]-1,y0+blue_line*6}})
                  end if
                  if w[cch-fc+1]<maxw*8 then
                     draw_line(9,{{x0+w[cch-fc+1]*6+1,y0+blue_line*6},{x0+6*8*maxw-1,y0+blue_line*6}})
                  end if
               end if
            elsif mbox=24 and loaded then
               change_cell()
            elsif mbox=25 and loaded then
               pick_char()
            end if
         end if
      end if
   elsif key=esc then
      exit
   end if
end while

if graphics_mode(-1) then
end if
