-- selector.ex: file selector
-- Jiri Babor
-- baborj@gns.cri.nz

-- Mouse pointer should be turned of when calling the file selector!

include font.e
include gadgets.e
include mouse.e
include ratbag.e
include file.e
include sort.e
include image.e
include wildcard.e

constant
   true=1,
   cr=13,
   esc=27

object d, f

sequence
   b,bd,curr_dir,mask,original_dir,s,si,sc

integer
   cf,tf,bf,
   font,n,x,y,
   original_drive,
   current_drive,
   ld,rd,sd,         -- selected drive
   key,pf,pd

global sequence selector_colors

-- defaults
selector_colors={15,7,8,1}  -- top-left edge, face, bottom-right edge, text
sc=selector_colors

font = fload("thin8x10.f")

function div(integer a,integer b)
   return  floor(a/b)
end function   -- div

function file_string(integer i)
   if d[i][2][1]='d' then
      s=" [dir] "
   elsif d[i][3]>999999 then
      s=sprintf("%4.1f M ",d[i][3]/1000000)
   else
      s=sprintf("%6d ",d[i][3])
   end if
   return s & d[i][1]
end function -- file_string

procedure show_file(integer i)
   integer c1,c2
   if i=cf then
      c1=sc[4] c2=sc[2]
   else
      c1=sc[2] c2=sc[4]
   end if
   fx=x+12
   fy=y+55+(i-tf)*14
   rect(c1,1,{x+10,fy-2},{x+170,fy+9})
   ftc=c2
   fprint(file_string(i))
end procedure  -- show_file

procedure show_files()
   -- erase current set
   rect(7,1,{x+6,y+51},{x+174,y+276})
   for i = tf to bf do
      show_file(i)
   end for
end procedure  -- show_files

procedure go_down()
   if cf<bf then
      cf=cf+1
      show_file(cf-1)
      show_file(cf)
   elsif bf<n then    -- display the next screenfull
      cf=cf+1
      tf=cf
      bf=tf+15
      if bf>n then
         bf=n
      end if
      show_files()
   end if
end procedure  -- go_down

procedure go_up()
   if cf>tf then
      cf=cf-1
      show_file(cf+1)
      show_file(cf)
   elsif tf>1 then
      cf=cf-1
      bf=cf
      tf=bf-15
      show_files()
   end if
end procedure  -- go_up

procedure page_up()
   if tf>1 then
      tf=tf-16
      cf=tf
      bf=tf+15
      show_files()
   end if
end procedure  -- page_up()

procedure page_down()
   if bf<n then
      tf=tf+16
      cf=tf
      bf=tf+15
      if bf>n then
         bf=n
      end if
      show_files()
   end if
end procedure  --  page_down

procedure show_current_dir()
   rect(7,1,{x+5,y+280},{x+175,y+294})
   ftc=sc[4]
   fx=x+10
   fy=y+283
   s = current_dir()
   if length(s)>20 then
      s = s[1..20]
   end if
   fprint(s)
end procedure  -- show_current_dir

function match_mask(object d)
   sequence n
   n={}
   if sequence(d) then
      for i = 1 to length(d) do
         if wildcard_file(mask,d[i][1]) then
            n = append(n,d[i])
         end if
      end for
      if length(n) then
         return n
      else
         return -1
      end if
   else
      return -1
   end if
end function   -- match_mask

procedure start()
   d = dir(curr_dir)
   d = match_mask(d)
   if sequence(d) then
      d=sort(d)
      n=length(d)
      cf=1        -- current file - hilited
      tf=1        -- (current) top file
      bf=tf+15
      if bf>n then
         bf=n
      end if
      show_files()
   else
      -- erase current set
      rect(7,1,{x+6,y+51},{x+174,y+276})
      fx=x+12
      fy=y+55
      ftc=sc[4]
      fprint("* No match found! *")
   end if
   show_current_dir()
end procedure  -- start

procedure show_current_drive()
   rect(7,1,{x+161,y+6},{x+174,y+21})
   ftc=sc[4]
   fx=x+164
   fy=y+10
   fprint({current_drive})
end procedure  -- show_current_drive

procedure show_header()
   fp[6]=1
   ftc=sc[4]
   fx=x+6
   fy=y+10
   fprint("m"&0&"ask")
   cs={sc[3],7,sc[1]}
   frame({x+40,y+5},{x+110,y+22})
   fx=x+45
   fprint(mask)
   fp[6]=1
   fx=x+117
   fprint("d"&0&"rive")
   frame({x+160,y+5},{x+175,y+22})
end procedure  -- show_header

procedure show_drive_letters()
   integer c1,c2
   rect(7,1,{x+6,y+6},{x+139,y+21})
   fy=y+10
   for i = ld to rd do
      fx = x+(i-ld)*13+10
      if i=sd then
         c1=sc[4] c2=sc[2]
      else
         c1=sc[2] c2=sc[4]
      end if
      ftc=c2
      rect(c1,1,{fx-2,fy-1},{fx+9,fy+9})
      fprint({i})
   end for
end procedure  -- show_drive_letters

procedure next_drive()
   if sd<'Z' then
      sd = sd+1
   end if
   if sd>rd then
      ld=ld+10
      rd=ld+9
      if rd>'Z' then rd='Z' end if
      show_drive_letters()
   else
      fx=x+(sd-ld-1)*13+10
      rect(7,1,{fx-2,fy-1},{fx+9,fy+9})
      ftc=sc[4]
      fprint({sd-1})
      fx=x+(sd-ld)*13+10
      rect(1,1,{fx-2,fy-1},{fx+9,fy+9})
      ftc=sc[2]
      fprint({sd})
   end if
end procedure  -- next_drive

procedure prev_drive()
   if sd>'A' then
      sd = sd-1
   end if
   if sd<ld then
      ld=ld-10
      rd=ld+9
      show_drive_letters()
   else
      fx=x+(sd-ld+1)*13+10
      rect(7,1,{fx-2,fy-1},{fx+9,fy+9})
      ftc=sc[4]
      fprint({sd+1})
      fx=x+(sd-ld)*13+10
      rect(1,1,{fx-2,fy-1},{fx+9,fy+9})
      ftc=sc[2]
      fprint({sd})
   end if
end procedure  -- prev_drive

procedure select_drive()
   -- erase existing top windows
   rect(7,1,{x+5,y+5},{x+175,y+22})
   cs={sc[3],7,sc[1]}
   frame({x+5,y+5},{x+140,y+22})
   cs = sc[1..3]
   ftc=sc[4]
   button({x+144,y+5},{x+159,y+22},{17})
   button({x+160,y+5},{x+175,y+22},{16})

   -- select display range
   ld='A'
   sd = current_drive
   while sd>ld+9 do
      ld=ld+10
   end while
   rd=ld+9
   if rd>'Z' then rd='Z' end if
   show_drive_letters()

   bd = {{{x+  5,y+ 27},{x+ 55,y+ 45}},   -- ok
         {{x+124,y+ 27},{x+175,y+ 45}},   -- cancel
         {{x+144,y+  5},{x+159,y+ 22}},   -- left arrow
         {{x+160,y+  5},{x+175,y+ 22}},   -- right arrow
         {{x+  5,y+  5},{x+140,y+ 22}}}   -- letter window

   while true do
      mouse_pointer(1)
      while true do
         mouse_event = get_mouse()
         key = get_key()
         if sequence(mouse_event) or key!=-1 then
            exit
         end if
      end while
      mouse_pointer(0)
      if sequence(mouse_event) then
         if mouse_event[1]=LEFT_DOWN then
            if mouse_box(bd) then
               if mbox=1 then     -- ok button
                  if sd!=current_drive then
                     current_drive=sd
                     system({current_drive,58},2)
                     curr_dir = current_dir()
                     start()
                  end if
                  exit
               elsif mbox=5 then  -- letter window
                  pd=div(mouse_event[2]-x-10,13)+ld
                  if pd<='Z' and pd>='A' then
                     sd=pd
                     if sd!=current_drive then
                        current_drive=sd
                        system({current_drive,58},2)
                        curr_dir = current_dir()
                        start()
                     end if
                     exit
                  end if
               elsif mbox=2 then  -- cancel button
                  exit
               elsif mbox=3 then  -- left button
                  prev_drive()
               elsif mbox=4 then  -- right button
                  next_drive()
               elsif mbox=5 then  -- letters
                  -- not yet
               end if
            end if
         end if
      elsif key=esc or key='n' or key='N' then  -- cancelled
         exit
      elsif key=cr or key='o' or key='O' then    -- o.k.
         if sd!=current_drive then
            current_drive=sd
            system({current_drive,58},2)
            curr_dir = current_dir()
            start()
         end if
         exit
      elsif key=331 then               -- left arrow
         prev_drive()
      elsif key=333 then               -- right arrow
         next_drive()
      end if
   end while
   -- erase & restore
   rect(7,1,{x+5,y+5},{x+175,y+22})
   show_header()
   show_current_drive()
end procedure  -- select_drive

procedure change_mask()
   fx=x+45
   fy=y+10
   fpc=7
   ftc=sc[4]
   mask=fgets(mask)
end procedure  -- change_mask

--------------------------------------------------------------------------------
global function selected(integer x0,integer y0,sequence mask0,integer path_flag)
   -- return:  -1 .. no match found
   --           0 .. cancelled ('cancel' button pressed)
   --           name of selected file
--------------------------------------------------------------------------------
   x=x0  y=y0  mask=mask0
   fselect(font)
   fp[1]=0
   si=save_image({x,y},{x+180,y+295})
   set_mouse_xy(x+66,y+38)
   cs = sc[1..3]
   frame({x,y},{x+180,y+295})
   cs={sc[3],7,sc[1]}
   frame({x+5,y+50},{x+175,y+277})
   show_header()
   -- buttons
   cs=sc[1..3]
   fp[6]=1
   ftc=sc[4]
   button({x+5,y+27},{x+55,y+45},"o"&0&"k")
   cs=sc[1..3]
   button({x+56,y+27},{x+72,y+45},{31})    -- down
   frame({x+73,y+27},{x+89,y+45})          -- page down
   fx=x+78
   fy=y+29
   fprint({31})
   fx=x+78
   fy=y+34
   fprint({31})
   frame({x+90,y+27},{x+106,y+45})         -- page up
   fx=x+95
   fy=y+29
   fprint({30})
   fx=x+95
   fy=y+34
   fprint({30})
   button({x+107,y+27},{x+123,y+45},{30})  -- up
   fp[6]=1
   button({x+124,y+27},{x+175,y+45},"n"&0&"ah")

   -- set mouse boxes
   b = {{{x+  5,y+ 27},{x+ 55,y+ 45}},    -- ok
        {{x+ 56,y+ 27},{x+ 72,y+ 45}},    -- down
        {{x+ 73,y+ 27},{x+ 89,y+ 45}},    -- page down
        {{x+ 90,y+ 27},{x+106,y+ 45}},    -- page up
        {{x+107,y+ 27},{x+123,y+ 45}},    -- up
        {{x+124,y+ 27},{x+175,y+ 45}},    -- cancel
        {{x+120,y+  5},{x+175,y+ 22}},    -- select drive
        {{x+  5,y+  5},{x+110,y+ 22}},    -- mask
        {{x+  5,y+ 53},{x+175,y+277}}}    -- file display

   original_dir = current_dir()
   curr_dir = original_dir
   original_drive = curr_dir[1]
   current_drive = curr_dir[1]
   show_current_drive()
   start()

   -- main loop
   while true do
      mouse_pointer(1)
      while true do
         mouse_event = get_mouse()
         key = get_key()
         if sequence(mouse_event) or key!=-1 then
            exit
         end if
      end while
      mouse_pointer(0)
      if sequence(mouse_event) then
         if mouse_event[1]=LEFT_DOWN then
            if mouse_box(b) then
               if mbox=1 then     -- o.k. button
                  if sequence(d) then
                     if d[cf][2][1]='d' then -- directory selected
                        system("cd " & d[cf][1], 2)
                        curr_dir = current_dir()
                        start()
                     else
                        f=d[cf][1]
                        if path_flag then
                           if curr_dir[length(curr_dir)]=92 then
                              f=curr_dir & f
                           else
                              f=curr_dir & 92 & f
                           end if
                        end if
                        exit
                     end if
                  else
                     f=0
                     exit
                  end if
               elsif mbox=9 then
                  pf=div(mouse_event[3]-y-53,14)+tf
                  if pf<=n and pf>0 then
                     cf=pf
                     if d[cf][2][1]='d' then -- directory selected
                        system("cd " & d[cf][1], 2)
                        curr_dir = current_dir()
                        start()
                     else
                        f=d[cf][1]
                        if path_flag then
                           if curr_dir[length(curr_dir)]=92 then
                              f=curr_dir & f
                           else
                              f=curr_dir & 92 & f
                           end if
                        end if
                        exit
                     end if
                  end if
               elsif mbox=2 then
                  go_down()
               elsif mbox=3 then
                  page_down()
               elsif mbox=4 then
                  page_up()
               elsif mbox=5 then
                  go_up()
               elsif mbox=6 then  -- cancel button
                  f=-1
                  exit
               elsif mbox=7 then
                  select_drive()
               elsif mbox=8 then
                  change_mask()
                  start()
               end if
            end if
         end if
      elsif key=esc or key='n' or key='N' then  -- cancelled
         f=-1
         exit
      elsif key='d' or key='D' then
         select_drive()
      elsif key='m' or key='M' then
         change_mask()
         start()
      elsif key=336 then
         go_down()
      elsif key=328 then
         go_up()
      elsif key=337 then
         page_down()
      elsif key=329 then
         page_up()
      elsif key=cr or key='o' or key='O' then   -- return
         if sequence(d) then
            if d[cf][2][1]='d' then -- directory selected
               system("cd " & d[cf][1], 2)
               curr_dir = current_dir()
               start()
            else
               f=d[cf][1]
               if path_flag then
                  if curr_dir[length(curr_dir)]=92 then
                     f=curr_dir & f
                  else
                     f=curr_dir & 92 & f
                  end if
               end if
               exit
            end if
         else
            f=0
            exit
         end if
      end if
   end while
   -- return to your original directory
   if original_drive!=current_drive then
      system({original_drive,58},2)
   end if
   if compare(original_dir,curr_dir) then
      system("cd " & original_dir, 2)
   end if
   -- restore the screen
   display_image({x,y},si)
   return f
end function   -- selected
