-- textmenu.e : simple text mode menu system

-- Jiri Babor
-- j.babor@gns.cri.nz

-- This simple menu system can be used without a mouse, but if you want to use
-- the mouse, remember the mouse pointer should be turned off before calling
-- the menu() function, and *both* mouse buttons should be monitored. Pressing
-- the right mouse button is equivalent to pressing the Esc key, i.e. to
-- descending one level down the menu tree.

include mouse.e
include ratbag.e
include image.e

constant
   true=1,
   false=0,
   cr=13,
   esc=27,
   left_arrow=331,
   right_arrow=333,
   up_arrow=328,
   down_arrow=336

type bool(integer x)
   return x=1 or x=0
end type

bool done

integer
   key,
   y,
   cl       -- current level

sequence
   mls,     -- menu level sequence
   x,       -- left edge co-ordinate at each level
   si,      -- sequence of saved images of areas under menus
   width,   -- menu item widths at each level
   m,       -- selection sequence returned by menu()
   mb,      -- sequence of mouse boxes at each level
   n        -- number of items in each level

global sequence menu_colors

-- defaults
menu_colors={1,7}  -- foreground, background

function div(integer a,integer b)
   return  floor(a/b)
end function   -- div

procedure set_current_level()
   integer w
   n=append(n, div(length(mls[cl]),2))   -- number of items in this level

   -- get width
   w=0
   for i = 1 to n[cl] do
      if length(mls[cl][2*i-1])>w then
         w=length(mls[cl][2*i-1])
      end if
   end for
   width=append(width,w+3)
   -- set up mouse box
   mb=append(mb,{{{x[cl],y},{x[cl]+width[cl]-1,y+n[cl]-1}}})
end procedure  --  set_current_level

procedure show_menu_item(integer item)
   sequence s
   if item=m[cl] then   -- invert face/text colors
      text_color(menu_colors[2])
      bk_color(menu_colors[1])
   else
      text_color(menu_colors[1])
      bk_color(menu_colors[2])
   end if
   s = 221 & mls[cl][2*item-1]
   while length(s)<width[cl] do
      s=s & " "
   end while
   if length(mls[cl][2*item]) then
      s[length(s)] = 16
   end if
   position(y+item-1,x[cl])
   puts(1,s)
end procedure  -- show_menu_item

procedure display_current_level()
   si = append(si, save_text_image({y,x[cl]},{y+n[cl]-1,x[cl]+width[cl]-1}))
   for i = 1 to n[cl] do
      show_menu_item(i)
   end for
end procedure  -- display_current_level

procedure down_level()
   display_text_image({y,x[cl]},si[cl])
   cl=cl-1
   x=x[1..cl]
   si=si[1..cl]
   m=m[1..cl]
   mls=mls[1..cl]
   n=n[1..cl]
   width=width[1..cl]
   mb=mb[1..cl]
   if cl=0 then   -- escaped without selection
      done=true
   end if
end procedure  -- down_level

procedure up_level()
   integer ci
   ci=m[cl]
   if length(mls[cl][2*ci]) then    -- set up and show submenu
      x=append(x, x[cl]+width[cl])
      mls=append(mls, mls[cl][2*ci])
      m=append(m,1)
      cl=cl+1
      set_current_level()
      display_current_level()
   end if
end procedure  -- up_level

procedure select()
   integer ci
   ci=m[cl]
   if length(mls[cl][2*ci]) then    -- set up and show submenu
      x=append(x, x[cl]+width[cl])
      mls=append(mls, mls[cl][2*ci])
      m=append(m,1)
      cl=cl+1
      set_current_level()
      display_current_level()
   else
      for i=cl to 1 by -1 do
         display_text_image({y,x[i]},si[i])
      end for
      done=true
   end if
end procedure  -- select

procedure mouse_selection()
   integer old_item
   old_item=m[cl]
   m[cl]=mouse_event[3]-y+1
   if m[cl]!=old_item then
      show_menu_item(old_item)
      show_menu_item(m[cl])
   end if
   select()
end procedure  -- mouse_selection

procedure next_item()
   if m[cl]<n[cl] then
      m[cl]=m[cl]+1
      show_menu_item(m[cl]-1)
      show_menu_item(m[cl])
   end if
end procedure  -- next_item

procedure prev_item()
   if m[cl]>1 then
      m[cl]=m[cl]-1
      show_menu_item(m[cl]+1)
      show_menu_item(m[cl])
   end if
end procedure  -- prev_item

global function menu(integer row, integer col, sequence s)
   if length(s) then
      -- init
      cursor(NO_CURSOR)
      n={}
      width={}
      x={col}
      y=row
      mls={s}           -- current menu sequence
      cl=1              -- current menu level
      m={cl}            -- current menu position: first item of root level menu
      si={}
      mb={}
      set_current_level()
      display_current_level()

      -- main loop
      done=false
      while not done do
         mouse_pointer(1)
         while true do
            mouse_event = get_text_mouse()
            key=get_key()
            if sequence(mouse_event) or  key!=-1 then
               exit
            end if
         end while
         mouse_pointer(0)
         if sequence(mouse_event) then
            if mouse_event[1]=RIGHT_DOWN then
               down_level()
            elsif mouse_event[1]=LEFT_DOWN then
               if mouse_box(mb[cl]) then
                  mouse_selection()
               end if
            end if
         elsif key=esc or key=left_arrow then
            down_level()
         elsif key=right_arrow then
            up_level()
         elsif key=cr then
            select()
         elsif key=down_arrow then
            next_item()
         elsif key=up_arrow then
            prev_item()
         end if
      end while
   else
      m={}
   end if
   cursor(THICK_UNDERLINE_CURSOR)
   return m
end function   -- menu
